-- LUALOCALS < ---------------------------------------------------------
local ItemStack, math, minetest, nodecore, pairs, string, vector
    = ItemStack, math, minetest, nodecore, pairs, string, vector
local math_exp, math_pi, math_random, string_format
    = math.exp, math.pi, math.random, string.format
-- LUALOCALS > ---------------------------------------------------------

------------------------------------------------------------------------
-- DATABASE

local modname = minetest.get_current_modname()
local modstore = minetest.get_mod_storage()

minetest.register_privilege(modname, {
		description = "Can \"borrow\" items from other players' offline inventory",
		give_to_admin = true,
		give_to_singleplayer = true
	})

local function privcheck(player)
	return player and player:is_player() and minetest.check_player_privs(player, modname)
end

local db = modstore:get_string("db")
if db == "" then db = nil end
db = db and minetest.deserialize(db)
db = db or {}

local function savedb()
	return modstore:set_string("db", minetest.serialize(db))
end

local function savestate(player, taken)
	local ent = {
		pos = player:get_pos(),
		inv = {},
		taken = taken
	}
	ent.pos.y = ent.pos.y + 1
	local inv = player:get_inventory()
	for i = 1, inv:get_size("main") do
		local stack = inv:get_stack("main", i)
		ent.inv[i] = stack:to_string()
	end
	ent.seen = minetest.get_gametime()
	db[player:get_player_name()] = ent
end

------------------------------------------------------------------------
-- INVENTORY OPEN CHECK

local invopen = {}

local function invopen_check(name, pos)
	local t = invopen[name]
	if t == nodecore.gametime then return t end
	if not t then return end

	for _, v in pairs(minetest.get_connected_players()) do
		local p = v:get_pos()
		p.y = p.y + v:get_properties().eye_height
		if vector.distance(p, pos) <= 5 then
			t = nodecore.gametime
			invopen[name] = t
			return t
		end
	end
	invopen[name] = nil
end

------------------------------------------------------------------------
-- MISC/UTILITY/COMMON

local function findentsforplayer(name, cb)
	for _, ent in pairs(minetest.luaentities) do
		if ent and ent.is_yctiwy and ent.pname == name then
			cb(ent)
		end
	end
end

local function areaunloaded(pos)
	local basepos = vector.floor(pos)
	for dx = -1, 1, 2 do
		for dy = -1, 1, 2 do
			for dz = -1, 1, 2 do
				if minetest.get_node({
						x = basepos.x + dx,
						y = basepos.y + dy,
						z = basepos.z + dz
					}).name == "ignore" then return true end
			end
		end
	end
end

local function possessive(s)
	return s .. (#s > 0 and s:sub(-1):lower() == "s" and "'" or "'s")
end

------------------------------------------------------------------------
-- INVENTORY ENTITIES

local cbs = 0.2

minetest.register_entity(modname .. ":slotent", {
		initial_properties = nodecore.stackentprops(),
		is_yctiwy = true,
		slot = 1,
		loose = 0,
		on_step = function(self, dtime)
			self.loose = self.loose / math_exp(dtime / 5)
			self.cktime = (self.cktime or 0) - dtime
			if self.cktime > 0 then return end
			self.cktime = 1

			local ent = db[self.pname]
			if minetest.get_player_by_name(self.pname)
			or not (ent and invopen_check(self.pname, ent.pos)) then
				return self.object:remove()
			end
			local stack = ent.inv and ent.inv[self.slot]
			if ent.taken and ent.taken[self.slot] then stack = "" end
			stack = ItemStack(stack or "")
			local def = minetest.registered_items[stack:get_name()] or {}
			if def.virtual_item then stack = ItemStack("") end
			local props = nodecore.stackentprops(stack)
			props.infotext = possessive(self.pname) .. " Inventory\n"
			.. nodecore.touchtip_stack(stack)
			props.visual_size.x = props.visual_size.x / 2
			props.visual_size.y = props.visual_size.y / 2
			if props.is_visible then
				props.collisionbox = {-cbs, -cbs, -cbs, cbs, cbs, cbs}
			end
			if not self.yawed then
				self.yawed = true
				self.object:set_yaw(math_random() * math_pi * 2)
			end
			self.object:set_properties(props)
		end,
		on_punch = function(self, puncher)
			if not privcheck(puncher) then return end
			self.loose = self.loose + 1
			if self.loose < 4 then return end
			local ent = db[self.pname]
			local stack = ent.inv and ent.inv[self.slot]
			if ent.taken and ent.taken[self.slot] then stack = "" end
			stack = ItemStack(stack or "")
			if stack:is_empty() then return end
			minetest.log(string_format("%s: %q steals %q slot %d item %q",
					modname, puncher:get_player_name(), self.pname,
					self.slot, stack:to_string()))
			ent.taken = ent.taken or {}
			ent.taken[self.slot] = true
			savedb()
			stack = puncher:get_inventory():add_item("main", stack)
			if not stack:is_empty() then
				nodecore.item_eject(self.object:get_pos(), stack)
			end
			return self.object:remove()
		end
	})

local s = 0.25
local offsets = {
	{x = -s, y = -s, z = -s},
	{x = s, y = -s, z = -s},
	{x = -s, y = s, z = -s},
	{x = s, y = s, z = -s},
	{x = -s, y = -s, z = s},
	{x = s, y = s, z = s},
	{x = -s, y = s, z = s},
	{x = s, y = -s, z = s}
}

local function spawninv(name, entry)
	entry = entry or db[name]
	if not entry then return end

	invopen[name] = nodecore.gametime
	if not invopen_check(name, entry.pos) then return end
	if areaunloaded(entry.pos) then
		invopen[name] = nil
		return
	end

	local exist = {}
	findentsforplayer(name, function(o)
			if o.slot then exist[o.slot] = true end
		end)
	for i = 1, #offsets do
		if not exist[i] and entry.inv[i] ~= "" then
			local p = vector.add(offsets[i], entry.pos)
			minetest.log(string_format("%s: adding slotent %d for %q at %s",
					modname, i, name, minetest.pos_to_string(p)))
			local o = minetest.add_entity(p, modname .. ":slotent")
			o = o and o:get_luaentity()
			if o then
				o.pname = name
				o.slot = i
			end
		end
	end
end

------------------------------------------------------------------------
-- MARKER ENTITY

minetest.register_entity(modname .. ":marker", {
		initial_properties = {
			visual = "sprite",
			textures = {"nc_player_wield_slot.png"},
			collisionbox = {-cbs, -cbs, -cbs, cbs, cbs, cbs},
			visual_size = {x = 0.4, y = 0.4},
			is_visible = true,
			static_save = false
		},
		is_yctiwy = true,
		compact = true,
		on_step = function(self, dtime)
			self.cktime = (self.cktime or 0) - dtime
			if self.cktime > 0 then return end
			self.cktime = 1

			local ent = db[self.pname]
			if minetest.get_player_by_name(self.pname)
			or (not minetest.player_exists(self.pname))
			or ent and invopen_check(self.pname, ent.pos) then
				return self.object:remove()
			end
		end,
		on_punch = function(self, puncher)
			if not privcheck(puncher) then return end
			self.punchcb()
			return self.object:remove()
		end
	})

local function spawnmarker(name, entry)
	entry = entry or db[name]
	if not entry then return end

	local any
	for k, v in pairs(entry.inv) do
		if not (entry.taken and entry.taken[k]) then
			local stack = ItemStack(v)
			if not (stack:is_empty() or stack:get_definition().virtual_item) then
				any = true
				break
			end
		end
	end
	if not any then return end

	if areaunloaded(entry.pos) then return end
	if invopen_check(name, entry.pos) then return end

	local exist
	findentsforplayer(name, function() exist = true end)
	if exist then return end

	minetest.log(string_format("%s: adding marker for %q at %s",
			modname, name, minetest.pos_to_string(entry.pos)))
	local o = minetest.add_entity(entry.pos, modname .. ":marker")
	if o then
		o:set_properties({
				infotext = possessive(name) .. " Inventory"
			})
		o = o:get_luaentity()
	end
	if o then
		o.pname = name
		o.punchcb = function()
			return spawninv(name, entry)
		end
	end
end

------------------------------------------------------------------------
-- PLAYER HOOKS

minetest.register_on_leaveplayer(function(player)
		savestate(player)
		return savedb()
	end)

minetest.register_on_shutdown(function()
		for _, pl in pairs(minetest.get_connected_players()) do
			savestate(pl)
		end
		return savedb()
	end)

minetest.register_on_joinplayer(function(player)
		local name = player:get_player_name()
		invopen[name] = nil
		local ent = db[name]
		if (not ent) or (not ent.taken) then return end
		local inv = player:get_inventory()
		for k in pairs(ent.taken) do
			inv:set_stack("main", k, "")
		end
		findentsforplayer(player:get_player_name(), function(o)
				o.object:remove()
			end)
	end)

------------------------------------------------------------------------
-- TIMER

local function dumpinv(name)
	local ent = db[name]
	if (not ent) or areaunloaded(ent.pos) then return end
	for k, v in pairs(ent.inv) do
		local stack = ItemStack(v)
		if not stack:is_empty() then
			if minetest.add_item(ent.pos, stack) then
				ent.inv[k] = ""
			else
				return
			end
		end
	end
	db[name] = nil
end

local timer = 0
minetest.register_globalstep(function(dtime)
		timer = timer - dtime
		if timer > 0 then return end
		timer = 3 + math_random() * 2

		local rollcall = {}
		for _, pl in pairs(minetest.get_connected_players()) do
			rollcall[pl:get_player_name()] = true
			savestate(pl)
		end

		for name, ent in pairs(db) do
			if not minetest.player_exists(name) then
				dumpinv(name)
			elseif not rollcall[name] then
				spawnmarker(name, ent)
			end
		end

		return savedb()
	end)
